/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.factsheet.wikipedia;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.Stack;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import org.xml.sax.Attributes;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;
import it.tidalwave.role.LocalizedDisplayable;
import it.tidalwave.role.MutableLocalizedDisplayable;
import it.tidalwave.role.spi.DefaultMutableDisplayable;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class WikipediaInfoExtractor
  {
    private final String scientificName;

    private final StringBuilder buffer = new StringBuilder();

    private final Stack<Status> stack = new Stack<Status>();

    private final MutableLocalizedDisplayable displayable = new DefaultMutableDisplayable("", "");

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private enum Status
      {
        IGNORE_ALL
          {
            public void processText (final @Nonnull WikipediaInfoExtractor textExtractor,
                                     final @Nonnull String text)
              {
              }

            public void openTag (final @Nonnull WikipediaInfoExtractor textExtractor,
                                 final @Nonnull String tag)
              {
              }

            public void closeTag (final @Nonnull WikipediaInfoExtractor textExtractor,
                                  final @Nonnull String tag)
              {
              }
          },

        IGNORE_TAG
          {
            public void openTag (final @Nonnull WikipediaInfoExtractor textExtractor,
                                 final @Nonnull String tag)
              {
              }

            public void closeTag (final @Nonnull WikipediaInfoExtractor textExtractor,
                                  final @Nonnull String tag)
              {
              }

            public void processText (final @Nonnull WikipediaInfoExtractor textExtractor,
                                     final @Nonnull String text)
              {
                textExtractor.buffer.append(text);
              }
          },

        APPEND_TEXT
          {
            public void openTag (final @Nonnull WikipediaInfoExtractor textExtractor,
                                 final @Nonnull String tag)
              {
                textExtractor.buffer.append("<").append(tag).append(">");
              }

            public void closeTag (final @Nonnull WikipediaInfoExtractor textExtractor,
                                  final @Nonnull String tag)
              {
                textExtractor.buffer.append("</").append(tag).append(">");
              }

            public void processText (final @Nonnull WikipediaInfoExtractor textExtractor,
                                     final @Nonnull String text)
              {
                textExtractor.buffer.append(text);
              }
          };

        public abstract void processText (@Nonnull WikipediaInfoExtractor textExtractor,
                                          @Nonnull String text);

        public abstract void openTag (@Nonnull WikipediaInfoExtractor textExtractor,
                                      @Nonnull String tag);

        public abstract void closeTag (@Nonnull WikipediaInfoExtractor textExtractor,
                                       @Nonnull String tag);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final EntityResolver entityResolver = new EntityResolver()
      {
        @Nonnull
        public InputSource resolveEntity (final @Nonnull String publicId, final @Nonnull String systemId)
          throws SAXException, IOException
          {
            if ("http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd".equals(systemId))
              {
                return new InputSource(WikipediaInfoExtractor.class.getResourceAsStream("xhtml1-transitional.dtd"));
              }

            return null;
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final DefaultHandler handler = new DefaultHandler()
      {
        private Status defaultStatus = Status.IGNORE_ALL;

        private List<String> ignoreTags = Arrays.asList("a", "span");
        private List<String> ignoreContents = Arrays.asList("table", "script");

        private String language = "";
        
        @Override
        public void startElement (final @Nonnull String nsURI,
                                  final @Nonnull String strippedName,
                                  final @Nonnull String tagName,
                                  final @Nonnull Attributes attributes)
          throws SAXException
          {
            if ("li".equalsIgnoreCase(strippedName))
              {
                final String clazz = attributes.getValue("class");
                language = ((clazz != null) && clazz.startsWith("interwiki-")) ? clazz.replace("interwiki-", "") : "";
              }
            else if ("a".equalsIgnoreCase(strippedName))
              {
                final String href = attributes.getValue("href");

                if (!"".equals(language))
                  {
                    try
                      {
                        final String value = URLDecoder.decode(href.replaceAll(".*/wiki/", "").replace('_', ' '), "UTF-8");
                        //
                        // Some languages use the scientific name in the URL
                        //
                        if (!scientificName.equals(value))
                          {
                            displayable.setDisplayName(value, new Locale(language));
                          }
                      }
                    catch (UnsupportedEncodingException e)
                      {
                      }
                  }
              }

            if ("body".equalsIgnoreCase(strippedName))
              {
                defaultStatus = Status.APPEND_TEXT;
                stack.push(Status.IGNORE_TAG);
              }
            else if ("div".equalsIgnoreCase(strippedName))
              {
                final String id = attributes.getValue("id");
//                  System.err.println("ATTR " + id);
                stack.push("bodyContent".equals(id) ? Status.IGNORE_TAG : Status.IGNORE_ALL);
              }
            else if (stack.peek() == Status.IGNORE_ALL)
              {
                stack.push(Status.IGNORE_ALL);
              }
            else if (ignoreTags.contains(tagName.toLowerCase()))
              {
                stack.push(Status.IGNORE_TAG);
              }
            else if (ignoreContents.contains(tagName.toLowerCase()))
              {
                stack.push(Status.IGNORE_ALL);
              }
            else
              {
                stack.push(defaultStatus);
              }

//              System.err.println("OPEN TAG: " + strippedName + " NEW STATUS: " + stack.peek());

            stack.peek().openTag(WikipediaInfoExtractor.this, strippedName);
          }

        @Override
        public void endElement (final @Nonnull String nsURI,
                                final @Nonnull String strippedName,
                                final @Nonnull String tagName)
          throws SAXException
          {
            stack.peek().closeTag(WikipediaInfoExtractor.this, strippedName);
            stack.pop();
//              System.err.println("CLOSE TAG: " + strippedName + " NEW STATUS: " + stack.peek());

            if ("body".equalsIgnoreCase(strippedName))
              {
                defaultStatus = Status.IGNORE_ALL;
              }
          }

        @Override
        public void characters (final @Nonnull char[] chars,
                                final @Nonnegative int start,
                                final @Nonnegative int length)
          {
            final String string = new String(chars, start, length);

            if (!string.trim().equals(""))
              {
                stack.peek().processText(WikipediaInfoExtractor.this, string);
              }
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public WikipediaInfoExtractor (final @Nonnull String scientificName)
      {
        this.scientificName = scientificName;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void process (final @Nonnull String path)
      throws SAXException, IOException
      {
        final XMLReader parser = createParser();
        parser.parse(path);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void process (final @Nonnull URL url)
      throws SAXException, IOException
      {
        final XMLReader parser = createParser();
        final InputStream is = url.openStream();
        parser.parse(new InputSource(is));
        is.close(); // FIXME: safe close
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public String getDescription()
      {
        return buffer.toString();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public LocalizedDisplayable getDisplayable()
      {
        return displayable;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private XMLReader createParser()
      throws SAXException
      {
        stack.push(Status.IGNORE_ALL);
        final XMLReader parser = XMLReaderFactory.createXMLReader();
        parser.setContentHandler(handler);
        parser.setEntityResolver(entityResolver);
        
        return parser;
      }
  }
